import inspect
from typing import Any, Dict

FromStore = Any


class ToStore(dict):
  pass


def update(func: Any, store: Dict[str, Any]) -> Any:
  stored_values = {}
  sig = inspect.signature(func)
  pars = sig.parameters
  new_pars = {}

  for item in pars.items():
    key = item[0]
    if item[1].annotation == FromStore:
      stored_values[key] = store[key]
    else:
      new_pars[key] = pars[key]

  def called(*args, **kwargs):
    res = func(*args, **kwargs, **stored_values)
    if type(res) == ToStore:
      res = {**store, **res}
    return res

  an = sig.return_annotation
  called.__signature__ = inspect.Signature(list(new_pars.values()),
                                           return_annotation=an)
  return called


def generator(param: float) -> ToStore:
  return ToStore(a=3. * param, b=5.)


def consumer_one(a: FromStore) -> ToStore:
  return ToStore(c=a + 2.)


def consumer_two(b: FromStore) -> ToStore:
  return ToStore(d=b + 5.)


def merger(c: FromStore, d: FromStore) -> ToStore:
  return ToStore(e=c + d)


if __name__ == "__main__":
  store = {}
  print(f'store before funcs: {store}')
  store = update(generator, store)(4.)
  print(f'store after generator: {store}')
  store = update(consumer_one, store)()
  print(f'store after consumer_one: {store}')
  store = update(consumer_two, store)()
  print(f'store after consumer_two: {store}')
  store = update(merger, store)()
  print(f'store after merger: {store}')
